@@ -12,64 +12,44 @@
 ##   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ##   GNU General Public License for more details.
 
-# $Id: simplexml.py,v 1.27 2005/04/30 07:20:27 snakeru Exp $
-
 """
 Simplexml module provides xmpppy library with all needed tools to handle XML
 nodes and XML streams. I'm personally using it in many other separate
 projects. It is designed to be as standalone as possible
 """
 
-from __future__ import unicode_literals
+from __future__ import annotations
+
+from typing import Dict
+from typing import List
+from typing import Optional
+from typing import Union
+from typing import Iterator
+from typing import Callable
+from typing import Any
 
-import sys
-import xml.parsers.expat
 import logging
+import xml.parsers.expat
+from xml.parsers.expat import ExpatError
+from copy import deepcopy
+
+from nbxmpp.const import NOT_ALLOWED_XML_CHARS
+
+
+Attrs = Dict[str, str]
+
 log = logging.getLogger('nbxmpp.simplexml')
 
-def XMLescape(txt):
+def XMLescape(text: str) -> str:
     """
-    Return provided string with symbols & < > " replaced by their respective XML
-    entities
+    Return escaped text
     """
-    # replace also FORM FEED and ESC, because they are not valid XML chars
-    return txt.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;").replace('"', "&quot;").replace('\x0C', "").replace('\x1B', "")
 
-ENCODING='utf-8'
+    for key, value in NOT_ALLOWED_XML_CHARS.items():
+        text = text.replace(key, value)
+    return text
 
-if sys.version_info[0] == 2:
-    def ustr(what):
-        """
-        Converts object "what" to unicode string using it's own __str__ method if
-        accessible or unicode method otherwise
-        """
-        if isinstance(what, unicode):
-            return what
-        try:
-            r = what.__str__()
-        except AttributeError:
-            r = unicode(what)
-        if not isinstance(r, unicode):
-            return unicode(r, ENCODING)
-        return r
-
-else:
-    def ustr(what):
-        """
-        Converts object "what" to unicode string using it's own __str__ method if
-        accessible or unicode method otherwise
-        """
-        if isinstance(what, str):
-            return what
-        try:
-            r = what.__str__()
-        except AttributeError:
-            r = str(what)
-        if not isinstance(r, str):
-            return str(r, ENCODING)
-        return r
-
-class Node(object):
+class Node:
     """
     Node class describes syntax of separate XML Node. It have a constructor that
     permits node creation from set of "namespace name", attributes and payload
@@ -90,17 +70,33 @@ class Node(object):
     on the classes tree).
     """
 
-    FORCE_NODE_RECREATION = 0
+    name: str
+    namespace: str
+    attrs: Attrs
+    data: List[str]
+    kids: List[Union[Node, str]]
+    parent: Optional[Node]
+    nsd: Dict[str, str]
+    nsp_cache: Dict[Any, Any]
 
-    def __init__(self, tag=None, attrs=None, payload=None, parent=None, nsp=None,
-                    node_built=False, node=None):
+    FORCE_NODE_RECREATION = False
+
+    def __init__(
+            self,
+            tag: Optional[str] = None,
+            attrs: Optional[Attrs] = None,
+            payload: Optional[Union[Node, str, List[Union[Node, str]]]] = None,
+            parent: Optional[Node] = None,
+            nsp: Optional[Dict[Any, Any]] = None,
+            node_built: bool = False,
+            node: Optional[Union[Node, Any]] = None) -> None:
         """
         Takes "tag" argument as the name of node (prepended by namespace, if
         needed and separated from it by a space), attrs dictionary as the set of
         arguments, payload list as the set of textual strings and child nodes
         that this node carries within itself and "parent" argument that is
-        another node that this one will be the child of. Also the __init__ can be
-        provided with "node" argument that is either a text string containing
+        another node that this one will be the child of. Also the __init__ can
+        be provided with "node" argument that is either a text string containing
         exactly one node or another Node instance to begin with. If both "node"
         and other arguments is provided then the node initially created as
         replica of "node" provided and then modified to be compliant with other
@@ -113,23 +109,36 @@ class Node(object):
                 node = NodeBuilder(node, self)
                 node_built = True
             else:
-                self.name, self.namespace, self.attrs, self.data, self.kids, self.parent, self.nsd = node.name, node.namespace, {}, [], [], node.parent, {}
+                self.name = node.name
+                self.namespace = node.namespace
+                self.attrs = {}
+                self.data = []
+                self.kids = []
+                self.parent = node.parent
+                self.nsd = {}
                 for key in node.attrs.keys():
                     self.attrs[key] = node.attrs[key]
                 for data in node.data:
                     self.data.append(data)
                 for kid in node.kids:
                     self.kids.append(kid)
-                for k, v in node.nsd.items():
-                    self.nsd[k] = v
+                for key, value in node.nsd.items():
+                    self.nsd[key] = value
         else:
-            self.name, self.namespace, self.attrs, self.data, self.kids, self.parent, self.nsd = 'tag', '', {}, [], [], None, {}
+            self.name = 'tag'
+            self.namespace = ''
+            self.attrs = {}
+            self.data = []
+            self.kids = []
+            self.parent = None
+            self.nsd = {}
         if parent:
             self.parent = parent
         self.nsp_cache = {}
         if nsp:
-            for k, v in nsp.items(): self.nsp_cache[k] = v
-           
+            for key, value in nsp.items():
+                self.nsp_cache[key] = value
+
         if attrs is not None:
             for attr, val in attrs.items():
                 if attr == 'xmlns':
@@ -154,9 +163,9 @@ class Node(object):
                 if isinstance(i, Node):
                     self.addChild(node=i)
                 else:
-                    self.data.append(ustr(i))
+                    self.data.append(str(i))
 
-    def lookup_nsp(self, pfx=''):
+    def lookup_nsp(self, pfx: str = '') -> str:
         ns = self.nsd.get(pfx, None)
         if ns is None:
             ns = self.nsp_cache.get(pfx, None)
@@ -168,10 +177,10 @@ class Node(object):
                 return 'http://www.gajim.org/xmlns/undeclared'
         return ns
 
-    def __str__(self, fancy=0):
+    def __str__(self, fancy: int = 0) -> str:
         """
-        Method used to dump node into textual representation. If "fancy" argument
-        is set to True produces indented output for readability
+        Method used to dump node into textual representation. If "fancy"
+        argument is set to True produces indented output for readability
         """
         s = (fancy-1) * 2 * ' ' + "<" + self.name
         if self.namespace:
@@ -179,7 +188,7 @@ class Node(object):
                 if 'xmlns' not in self.attrs:
                     s += ' xmlns="%s"' % self.namespace
         for key in self.attrs.keys():
-            val = ustr(self.attrs[key])
+            val = str(self.attrs[key])
             s += ' %s="%s"' % (key, XMLescape(val))
 
         s += ">"
@@ -213,7 +222,12 @@ class Node(object):
                 s += "\n"
         return s
 
-    def addChild(self, name=None, attrs=None, payload=None, namespace=None, node=None):
+    def addChild(self,
+                 name: Optional[str] = None,
+                 attrs: Optional[Attrs] = None,
+                 payload: Optional[List[Any]] = None,
+                 namespace: Optional[str] = None,
+                 node: Optional[Node] = None) -> Node:
         """
         If "node" argument is provided, adds it as child node. Else creates new
         node from the other arguments' values and adds it as well
@@ -234,25 +248,27 @@ class Node(object):
         self.kids.append(newnode)
         return newnode
 
-    def addData(self, data):
+    def addData(self, data: Any) -> None:
         """
         Add some CDATA to node
         """
-        self.data.append(ustr(data))
+        self.data.append(str(data))
 
-    def clearData(self):
+    def clearData(self) -> None:
         """
         Remove all CDATA from the node
         """
         self.data = []
 
-    def delAttr(self, key):
+    def delAttr(self, key: str) -> None:
         """
         Delete an attribute "key"
         """
         del self.attrs[key]
 
-    def delChild(self, node, attrs=None):
+    def delChild(self,
+                 node: Union[Node, str],
+                 attrs: Optional[Attrs] = None) -> Optional[Node]:
         """
         Delete the "node" from the node's childs list, if "node" is an instance.
         Else delete the first node that have specified name and (optionally)
@@ -260,58 +276,61 @@ class Node(object):
         """
         if not isinstance(node, Node):
             node = self.getTag(node, attrs)
+        assert isinstance(node, Node)
         self.kids.remove(node)
         return node
 
-    def getAttrs(self):
+    def getAttrs(self, copy: bool = False) -> Attrs:
         """
         Return all node's attributes as dictionary
         """
+        if copy:
+            return deepcopy(self.attrs)
         return self.attrs
 
-    def getAttr(self, key):
+    def getAttr(self, key: str) -> Optional[str]:
         """
         Return value of specified attribute
         """
         return self.attrs.get(key)
 
-    def getChildren(self):
+    def getChildren(self) -> List[Union[Node, str]]:
         """
         Return all node's child nodes as list
         """
         return self.kids
 
-    def getData(self):
+    def getData(self) -> str:
         """
         Return all node CDATA as string (concatenated)
         """
         return ''.join(self.data)
 
-    def getName(self):
+    def getName(self) -> str:
         """
         Return the name of node
         """
         return self.name
 
-    def getNamespace(self):
+    def getNamespace(self) -> str:
         """
         Return the namespace of node
         """
         return self.namespace
 
-    def getParent(self):
+    def getParent(self) -> Optional[Node]:
         """
         Returns the parent of node (if present)
         """
         return self.parent
 
-    def getPayload(self):
+    def getPayload(self) -> List[Union[Node, str]]:
         """
         Return the payload of node i.e. list of child nodes and CDATA entries.
         F.e. for "<node>text1<nodea/><nodeb/> text2</node>" will be returned
         list: ['text1', <nodea instance>, <nodeb instance>, ' text2']
         """
-        ret = []
+        ret: List[Union[Node, str]] = []
         for i in range(len(self.kids)+len(self.data)+1):
             try:
                 if self.data[i]:
@@ -324,33 +343,45 @@ class Node(object):
                 pass
         return ret
 
-    def getTag(self, name, attrs=None, namespace=None):
+    def getTag(self,
+               name: str,
+               attrs: Optional[Attrs] = None,
+               namespace: Optional[str] = None) -> Optional[Node]:
         """
         Filter all child nodes using specified arguments as filter. Return the
         first found or None if not found
         """
-        return self.getTags(name, attrs, namespace, one=1)
+        tag = self.getTags(name, attrs, namespace, one=True)
+        assert not isinstance(tag, list)
+        return tag
 
-    def getTagAttr(self, tag, attr, namespace=None):
+    def getTagAttr(self,
+                   tag: str,
+                   attr: str,
+                   namespace: Optional[str] = None) -> Optional[str]:
         """
         Return attribute value of the child with specified name (or None if no
         such attribute)
         """
-        try:
-            return self.getTag(tag, namespace=namespace).attrs[attr]
-        except:
+        node = self.getTag(tag, namespace=namespace)
+        if node is None:
             return None
+        return node.getAttr(attr)
 
-    def getTagData(self, tag):
+    def getTagData(self, tag: str) -> Optional[str]:
         """
         Return cocatenated CDATA of the child with specified name
         """
-        try:
-            return self.getTag(tag).getData()
-        except Exception:
+        node = self.getTag(tag)
+        if node is None:
             return None
+        return node.getData()
 
-    def getTags(self, name, attrs=None, namespace=None, one=0):
+    def getTags(self,
+                name: str,
+                attrs: Optional[Attrs] = None,
+                namespace: Optional[str] = None,
+                one: bool = False) -> Union[List[Node], Node, None]:
         """
         Filter all child nodes using specified arguments as filter. Returns the
         list of nodes found
@@ -363,7 +394,7 @@ class Node(object):
                 if attrs is None:
                     attrs = {}
                 for key in attrs.keys():
-                    if key not in node.attrs or node.attrs[key]!=attrs[key]:
+                    if key not in node.attrs or node.attrs[key] != attrs[key]:
                         break
                 else:
                     nodes.append(node)
@@ -371,8 +402,12 @@ class Node(object):
                 return nodes[0]
         if not one:
             return nodes
+        return None
 
-    def iterTags(self, name, attrs=None, namespace=None):
+    def iterTags(self,
+                 name: str,
+                 attrs: Optional[Attrs] = None,
+                 namespace: Optional[str] = None) -> Iterator[Node]:
         """
         Iterate over all children using specified arguments as filter
         """
@@ -389,39 +424,41 @@ class Node(object):
                 else:
                     yield node
 
-    def setAttr(self, key, val):
+    def setAttr(self, key: str, val: str) -> None:
         """
         Set attribute "key" with the value "val"
         """
         self.attrs[key] = val
 
-    def setData(self, data):
+    def setData(self, data: Any) -> None:
         """
         Set node's CDATA to provided string. Resets all previous CDATA!
         """
-        self.data = [ustr(data)]
+        self.data = [str(data)]
 
-    def setName(self, val):
+    def setName(self, val: str) -> None:
         """
         Change the node name
         """
         self.name = val
 
-    def setNamespace(self, namespace):
+    def setNamespace(self, namespace: str) -> None:
         """
         Changes the node namespace
         """
         self.namespace = namespace
 
-    def setParent(self, node):
+    def setParent(self, node: Node) -> None:
         """
-        Set node's parent to "node". WARNING: do not checks if the parent already
-        present and not removes the node from the list of childs of previous
-        parent
+        Set node's parent to "node". WARNING: do not checks if the parent
+        already present and not removes the node from the list of childs of
+        previous parent
         """
         self.parent = node
 
-    def setPayload(self, payload, add=0):
+    def setPayload(self,
+                   payload: Union[List[Union[Node, str]], Node, str],
+                   add: bool = False) -> None:
         """
         Set node payload according to the list specified. WARNING: completely
         replaces all node's previous content. If you wish just to add child or
@@ -434,18 +471,24 @@ class Node(object):
         else:
             self.kids = payload
 
-    def setTag(self, name, attrs=None, namespace=None):
+    def setTag(self,
+               name: str,
+               attrs: Optional[Attrs] = None,
+               namespace: Optional[str] = None) -> Node:
         """
         Same as getTag but if the node with specified namespace/attributes not
         found, creates such node and returns it
         """
-        node = self.getTags(name, attrs, namespace=namespace, one=1)
+        node = self.getTags(name, attrs, namespace=namespace, one=True)
         if node:
             return node
-        else:
-            return self.addChild(name, attrs, namespace=namespace)
+        return self.addChild(name, attrs, namespace=namespace)
 
-    def setTagAttr(self, tag, attr, val, namespace=None):
+    def setTagAttr(self,
+                   tag: str,
+                   attr: str,
+                   val: str,
+                   namespace: Optional[str] = None) -> None:
         """
         Create new node (if not already present) with name "tag" and set it's
         attribute "attr" to value "val"
@@ -455,47 +498,59 @@ class Node(object):
         except Exception:
             self.addChild(tag, namespace=namespace, attrs={attr: val})
 
-    def setTagData(self, tag, val, attrs=None):
+    def setTagData(self,
+                   tag: str,
+                   val: str,
+                   attrs: Optional[Attrs] = None) -> None:
         """
         Creates new node (if not already present) with name "tag" and
         (optionally) attributes "attrs" and sets it's CDATA to string "val"
         """
         try:
-            self.getTag(tag, attrs).setData(ustr(val))
+            self.getTag(tag, attrs).setData(str(val))
         except Exception:
-            self.addChild(tag, attrs, payload = [ustr(val)])
+            self.addChild(tag, attrs, payload = [str(val)])
 
-    def has_attr(self, key):
+    def getXmlLang(self) -> Optional[str]:
+        lang = self.attrs.get('xml:lang')
+        if lang is not None:
+            return lang
+
+        if self.parent is not None:
+            return self.parent.getXmlLang()
+        return None
+
+    def has_attr(self, key: str) -> bool:
         """
         Check if node have attribute "key"
         """
         return key in self.attrs
 
-    def __getitem__(self, item):
+    def __getitem__(self, item: str) -> Optional[str]:
         """
         Return node's attribute "item" value
         """
         return self.getAttr(item)
 
-    def __setitem__(self, item, val):
+    def __setitem__(self, item: str, val: str) -> None:
         """
         Set node's attribute "item" value
         """
-        return self.setAttr(item, val)
+        self.setAttr(item, val)
 
-    def __delitem__(self, item):
+    def __delitem__(self, item: str) -> None:
         """
         Delete node's attribute "item"
         """
-        return self.delAttr(item)
+        self.delAttr(item)
 
-    def __contains__(self, item):
+    def __contains__(self, item: str) -> bool:
         """
         Check if node has attribute "item"
         """
         return self.has_attr(item)
 
-    def __getattr__(self, attr):
+    def __getattr__(self, attr: str) -> Union['T', 'NT']:
         """
         Reduce memory usage caused by T/NT classes - use memory only when needed
         """
@@ -508,7 +563,7 @@ class Node(object):
         raise AttributeError
 
 
-class T(object):
+class T:
     """
     Auxiliary class used to quick access to node's child nodes
     """
@@ -522,8 +577,8 @@ class T(object):
     def __setattr__(self, attr, val):
         if isinstance(val, Node):
             Node.__init__(self.node.setTag(attr), node=val)
-        else:
-            return self.node.setTagData(attr, val)
+            return None
+        return self.node.setTagData(attr, val)
 
     def __delattr__(self, attr):
         return self.node.delChild(attr)
@@ -539,11 +594,11 @@ class NT(T):
     def __setattr__(self, attr, val):
         if isinstance(val, Node):
             self.node.addChild(attr, node=val)
-        else:
-            return self.node.addChild(attr, payload=[val])
+            return None
+        return self.node.addChild(attr, payload=[val])
 
 
-class NodeBuilder(object):
+class NodeBuilder:
     """
     Builds a Node class minidom from data parsed to it. This class used for two
     purposes:
@@ -557,7 +612,26 @@ class NodeBuilder(object):
     XML handler
     """
 
-    def __init__(self, data=None, initial_node=None):
+    _parser: Any
+    Parse: Callable[[str, bool], None]
+    __depth: int
+    __last_depth: int
+    __max_depth: int
+    _dispatch_depth: int
+    _document_attrs: Optional[Attrs]
+    _document_nsp: Optional[Dict[str, str]]
+    _mini_dom: Optional[Node]
+    last_is_data: bool
+    _ptr: Optional[Node]
+    data_buffer: Optional[List[str]]
+    streamError: str
+    _is_stream: bool
+
+    def __init__(self,
+                 data: Optional[str] = None,
+                 initial_node: Optional[Node] = None,
+                 dispatch_depth: int = 1,
+                 finished: bool = True) -> None:
         """
         Take two optional parameters: "data" and "initial_node"
 
@@ -566,36 +640,44 @@ class NodeBuilder(object):
         about it as of "node upgrade". "data" (if provided) feeded to parser
         immidiatedly after instance init.
         """
-        log.debug("Preparing to handle incoming XML stream.")
         self._parser = xml.parsers.expat.ParserCreate()
+        self._parser.UseForeignDTD(False)
         self._parser.StartElementHandler = self.starttag
         self._parser.EndElementHandler = self.endtag
         self._parser.StartNamespaceDeclHandler = self.handle_namespace_start
         self._parser.CharacterDataHandler = self.handle_cdata
+        self._parser.StartDoctypeDeclHandler = self.handle_invalid_xmpp_element
+        self._parser.EntityDeclHandler = self.handle_invalid_xmpp_element
+        self._parser.CommentHandler = self.handle_invalid_xmpp_element
+        self._parser.ExternalEntityRefHandler = self.handle_invalid_xmpp_element
+        self._parser.AttlistDeclHandler = self.handle_invalid_xmpp_element
+        self._parser.ProcessingInstructionHandler = \
+            self.handle_invalid_xmpp_element
         self._parser.buffer_text = True
         self.Parse = self._parser.Parse
 
         self.__depth = 0
         self.__last_depth = 0
         self.__max_depth = 0
-        self._dispatch_depth = 1
+        self._dispatch_depth = dispatch_depth
         self._document_attrs = None
         self._document_nsp = None
-        self._mini_dom=initial_node
-        self.last_is_data = 1
-        self._ptr=None
+        self._mini_dom = initial_node
+        self.last_is_data = True
+        self._ptr = None
         self.data_buffer = None
         self.streamError = ''
+        self._is_stream = not finished
         if data:
-            self._parser.Parse(data, 1)
+            self._parser.Parse(data, finished)
 
-    def check_data_buffer(self):
+    def check_data_buffer(self) -> None:
         if self.data_buffer:
             self._ptr.data.append(''.join(self.data_buffer))
             del self.data_buffer[:]
             self.data_buffer = None
 
-    def destroy(self):
+    def destroy(self) -> None:
         """
         Method used to allow class instance to be garbage-collected
         """
@@ -605,21 +687,32 @@ class NodeBuilder(object):
         self._parser.CharacterDataHandler = None
         self._parser.StartNamespaceDeclHandler = None
 
-    def starttag(self, tag, attrs):
+    def starttag(self, tag: str, attrs: Attrs) -> None:
         """
         XML Parser callback. Used internally
         """
         self.check_data_buffer()
         self._inc_depth()
-        log.debug("STARTTAG.. DEPTH -> %i , tag -> %s, attrs -> %s" % (self.__depth, tag, attrs))
+        log.debug("STARTTAG.. DEPTH -> %i , tag -> %s, attrs -> %s",
+                  self.__depth, tag, attrs)
         if self.__depth == self._dispatch_depth:
-            if not self._mini_dom :
-                self._mini_dom = Node(tag=tag, attrs=attrs, nsp = self._document_nsp, node_built=True)
+            if not self._mini_dom:
+                self._mini_dom = Node(tag=tag,
+                                      attrs=attrs,
+                                      nsp=self._document_nsp,
+                                      node_built=True)
             else:
-                Node.__init__(self._mini_dom, tag=tag, attrs=attrs, nsp = self._document_nsp, node_built=True)
+                Node.__init__(self._mini_dom,
+                              tag=tag,
+                              attrs=attrs,
+                              nsp=self._document_nsp,
+                              node_built=True)
             self._ptr = self._mini_dom
         elif self.__depth > self._dispatch_depth:
-            self._ptr.kids.append(Node(tag=tag, parent=self._ptr, attrs=attrs, node_built=True))
+            self._ptr.kids.append(Node(tag=tag,
+                                       parent=self._ptr,
+                                       attrs=attrs,
+                                       node_built=True))
             self._ptr = self._ptr.kids[-1]
         if self.__depth == 1:
             self._document_attrs = {}
@@ -632,24 +725,34 @@ class NodeBuilder(object):
                     self._document_nsp[attr[6:]] = val
                 else:
                     self._document_attrs[attr] = val
-            ns = self._document_nsp.get(nsp, 'http://www.gajim.org/xmlns/undeclared-root')
+            ns = self._document_nsp.get(
+                nsp, 'http://www.gajim.org/xmlns/undeclared-root')
             try:
-                header = Node(tag=tag, attrs=attrs,
+                header = Node(tag=tag,
+                              attrs=attrs,
                               nsp=self._document_nsp, node_built=True)
                 self.dispatch(header)
-                self.stream_header_received(ns, name, attrs)
-            except ValueError as e:
+                self._check_stream_start(ns, name)
+            except ValueError as error:
                 self._document_attrs = None
-                raise ValueError(str(e))
+                raise ValueError(str(error))
         if not self.last_is_data and self._ptr.parent:
             self._ptr.parent.data.append('')
-        self.last_is_data = 0
+        self.last_is_data = False
 
-    def endtag(self, tag ):
+    def _check_stream_start(self, ns: str, tag: str) -> None:
+        if self._is_stream:
+            if ns != 'http://etherx.jabber.org/streams' or tag != 'stream':
+                raise ValueError('Incorrect stream start: (%s,%s). Terminating.'
+                                 % (tag, ns))
+        else:
+            self.stream_header_received()
+
+    def endtag(self, tag: str) -> None:
         """
         XML Parser callback. Used internally
         """
-        log.debug("DEPTH -> %i , tag -> %s" % (self.__depth, tag))
+        log.debug("DEPTH -> %i , tag -> %s", self.__depth, tag)
         self.check_data_buffer()
         if self.__depth == self._dispatch_depth:
             if self._mini_dom.getName() == 'error':
@@ -664,78 +767,82 @@ class NodeBuilder(object):
         else:
             log.debug("Got higher than dispatch level. Stream terminated?")
         self._dec_depth()
-        self.last_is_data = 0
-        if self.__depth == 0: self.stream_footer_received()
+        self.last_is_data = False
+        if self.__depth == 0:
+            self.stream_footer_received()
 
-    def handle_cdata(self, data):
+    def handle_cdata(self, data: str) -> None:
         if self.last_is_data:
             if self.data_buffer:
                 self.data_buffer.append(data)
         elif self._ptr:
             self.data_buffer = [data]
-            self.last_is_data = 1
+            self.last_is_data = True
 
-    def handle_namespace_start(self, prefix, uri):
+    @staticmethod
+    def handle_invalid_xmpp_element(*args: Any) -> None:
+        raise ExpatError('Found invalid xmpp stream element: %s' % str(args))
+
+    def handle_namespace_start(self, _prefix: str, _uri: str) -> None:
         """
         XML Parser callback. Used internally
         """
         self.check_data_buffer()
 
-    def getDom(self):
+    def getDom(self) -> Optional[Node]:
         """
         Return just built Node
         """
         self.check_data_buffer()
         return self._mini_dom
 
-    def dispatch(self, stanza):
+    def dispatch(self, stanza: Any) -> None:
         """
         Get called when the NodeBuilder reaches some level of depth on it's way
         up with the built node as argument. Can be redefined to convert incoming
         XML stanzas to program events
         """
-        pass
 
-    def stream_header_received(self, ns, tag, attrs):
+    def stream_header_received(self) -> None:
         """
         Method called when stream just opened
         """
         self.check_data_buffer()
 
-    def stream_footer_received(self):
+    def stream_footer_received(self) -> None:
         """
         Method called when stream just closed
         """
         self.check_data_buffer()
 
-    def has_received_endtag(self, level=0):
+    def has_received_endtag(self, level: int = 0) -> bool:
         """
         Return True if at least one end tag was seen (at level)
         """
-        return self.__depth <= level and self.__max_depth > level
+        return self.__depth <= level < self.__max_depth
 
-    def _inc_depth(self):
+    def _inc_depth(self) -> None:
         self.__last_depth = self.__depth
         self.__depth += 1
         self.__max_depth = max(self.__depth, self.__max_depth)
 
-    def _dec_depth(self):
+    def _dec_depth(self) -> None:
         self.__last_depth = self.__depth
         self.__depth -= 1
 
-def XML2Node(xml):
+def XML2Node(xml_str: str) -> Optional[Node]:
     """
     Convert supplied textual string into XML node. Handy f.e. for reading
     configuration file. Raises xml.parser.expat.parsererror if provided string
     is not well-formed XML
     """
-    return NodeBuilder(xml).getDom()
+    return NodeBuilder(xml_str).getDom()
 
-def BadXML2Node(xml):
+def BadXML2Node(xml_str: str) -> Optional[Node]:
     """
     Convert supplied textual string into XML node. Survives if xml data is
     cutted half way round. I.e. "<html>some text <br>some more text". Will raise
     xml.parser.expat.parsererror on misplaced tags though. F.e. "<b>some text
     <br>some more text</b>" will not work
     """
-    return NodeBuilder(xml).getDom()
+    return NodeBuilder(xml_str).getDom()
